@@ -77,19 +77,34 @@ static void key_mnt_params_init(void)
 
 static void mnts_cleanup(void *arg)
 {
-	struct mnt_list *mnts = (struct mnt_list *) arg;
-	tree_free_mnt_tree(mnts);
-	return;
+	struct list_head *mnts = (struct list_head *) arg;
+	mnts_put_expire_list(mnts);
 }
 
 int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
+	struct mapent_cache *mc = me->mc;
 	char buf[MAX_ERR_BUF];
 	int ioctlfd = -1, rv, left, retries;
+	char key[PATH_MAX + 1];
+	struct mapent *tmp;
 	int opened = 0;
 
-	left = umount_multi(ap, me->key, 0);
+	if (me->len > PATH_MAX) {
+		error(ap->logopt, "path too long");
+		return 1;
+	}
+	strcpy(key, me->key);
+
+	cache_unlock(mc);
+	left = umount_multi(ap, key, 0);
+	cache_readlock(mc);
+	tmp = cache_lookup_distinct(mc, key);
+	if (tmp != me) {
+		error(ap->logopt, "key %s no longer in mapent cache", key);
+		return -1;
+	}
 	if (left) {
 		warn(ap->logopt, "could not unmount %d dirs under %s",
 		     left, me->key);
@@ -106,7 +121,9 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 		}
 		ioctlfd = me->ioctlfd;
 	} else {
-		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
+		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
+		if (ioctlfd == -1)
+			return 1;
 		opened = 1;
 	}
 
@@ -151,7 +168,7 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 200000000};
+		struct timespec tm = {0, 50000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -214,6 +231,7 @@ int umount_autofs_direct(struct autofs_point *ap)
 		mc = map->mc;
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		cache_readlock(mc);
+restart:
 		me = cache_enumerate(mc, NULL);
 		while (me) {
 			int error;
@@ -231,6 +249,9 @@ int umount_autofs_direct(struct autofs_point *ap)
 			 * failed umount.
 			 */
 			error = do_umount_autofs_direct(ap, me);
+			/* cache became invalid, restart */
+			if (error == -1)
+				goto restart;
 			if (!error)
 				goto done;
 
@@ -244,8 +265,6 @@ done:
 	}
 	pthread_cleanup_pop(1);
 
-	close(ap->state_pipe[0]);
-	close(ap->state_pipe[1]);
 	if (ap->pipefd >= 0)
 		close(ap->pipefd);
 	if (ap->kpipefd >= 0) {
@@ -286,15 +305,21 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 		if (ret == 0)
 			return -1;
 	} else {
-		if (ap->state == ST_READMAP && is_mounted(me->key, MNTS_ALL)) {
+		/* I don't remember why this is here for the force
+		 * unlink case. I don't think it should be but I may
+		 * have done it for a reason so keep it for the unlink
+		 * and continue case but not for the unlink and exit
+		 * case.
+		 */
+		if (!(do_force_unlink & UNLINK_AND_EXIT) &&
+		    ap->state == ST_READMAP && is_mounted(me->key, MNTS_ALL)) {
 			time_t tout = get_exp_timeout(ap, me->source);
 			int save_ioctlfd, ioctlfd;
 
 			save_ioctlfd = ioctlfd = me->ioctlfd;
 
 			if (ioctlfd == -1)
-				ops->open(ap->logopt,
-					  &ioctlfd, me->dev, me->key);
+				ioctlfd = open_ioctlfd(ap, me->key, me->dev);
 
 			if (ioctlfd < 0) {
 				error(ap->logopt,
@@ -311,7 +336,7 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 			return 0;
 		}
 
-		ret = unlink_mount_tree(ap, ap->path);
+		ret = unlink_mount_tree(ap, me->key);
 		if (!ret) {
 			error(ap->logopt,
 			     "already mounted as other than autofs "
@@ -319,6 +344,9 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 			goto out_err;
 		}
 
+		if (do_force_unlink & UNLINK_AND_EXIT)
+			return -1;
+
 		if (me->ioctlfd != -1) {
 			error(ap->logopt, "active direct mount %s", me->key);
 			return -1;
@@ -348,29 +376,10 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path, ap->kpipefd, str_direct);
+		mp->options = make_options_string(ap->path,
+				ap->kpipefd, str_direct, ap->flags);
 		if (!mp->options)
 			return 0;
-
-		if ((ap->flags & MOUNT_FLAG_STRICTEXPIRE) &&
-		    ((get_kver_major() == 5 && get_kver_minor() > 3) ||
-		     (get_kver_major() > 5))) {
-			char *tmp = realloc(mp->options, strlen(mp->options) + 12);
-			if (tmp) {
-				strcat(tmp, ",strictexpire");
-				mp->options = tmp;
-			}
-		}
-
-		if ((ap->flags & MOUNT_FLAG_IGNORE) &&
-		    ((get_kver_major() == 5 && get_kver_minor() > 4) ||
-		     (get_kver_major() > 5))) {
-			char *tmp = realloc(mp->options, strlen(mp->options) + 7);
-			if (tmp) {
-				strcat(tmp, ",ignore");
-				mp->options = tmp;
-			}
-		}
 	}
 
 	/* In case the directory doesn't exist, try to mkdir it */
@@ -402,11 +411,13 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 		      "failed to stat direct mount trigger %s", me->key);
 		goto out_umount;
 	}
+	me->dev = st.st_dev;
+	me->ino = st.st_ino;
 
 	if (ap->mode && (err = chmod(me->key, ap->mode)))
 		warn(ap->logopt, "failed to change mode of %s", me->key);
 
-	ops->open(ap->logopt, &ioctlfd, st.st_dev, me->key);
+	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
 	if (ioctlfd < 0) {
 		crit(ap->logopt, "failed to create ioctl fd for %s", me->key);
 		goto out_umount;
@@ -414,7 +425,7 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 
 	ops->timeout(ap->logopt, ioctlfd, timeout);
 	notify_mount_result(ap, me->key, timeout, str_direct);
-	cache_set_ino_index(me->mc, me->key, st.st_dev, st.st_ino);
+	cache_set_ino_index(me->mc, me);
 	ops->close(ap->logopt, ioctlfd);
 
 	debug(ap->logopt, "mounted trigger %s", me->key);
@@ -454,8 +465,6 @@ int mount_autofs_direct(struct autofs_point *ap)
 	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 	master_source_readlock(ap->entry);
 	nc = ap->entry->master->nc;
-	cache_readlock(nc);
-	pthread_cleanup_push(cache_lock_cleanup, nc);
 	map = ap->entry->maps;
 	while (map) {
 		time_t timeout;
@@ -474,9 +483,13 @@ int mount_autofs_direct(struct autofs_point *ap)
 		pthread_cleanup_push(cache_lock_cleanup, mc);
 		me = cache_enumerate(mc, NULL);
 		while (me) {
+			cache_writelock(nc);
 			ne = cache_lookup_distinct(nc, me->key);
 			if (ne) {
-				if (map->master_line < ne->age) {
+				unsigned int ne_age = ne->age;
+
+				cache_unlock(nc);
+				if (map->master_line < ne_age) {
 					/* TODO: check return, locking me */
 					do_mount_autofs_direct(ap, me, timeout);
 				}
@@ -485,13 +498,14 @@ int mount_autofs_direct(struct autofs_point *ap)
 			}
 
 			nested = cache_partial_match(nc, me->key);
-			if (nested) {
+			if (!nested)
+				cache_unlock(nc);
+			else {
+				cache_delete(nc, nested->key);
+				cache_unlock(nc);
 				error(ap->logopt,
 				   "removing invalid nested null entry %s",
 				   nested->key);
-				nested = cache_partial_match(nc, me->key);
-				if (nested)
-					cache_delete(nc, nested->key);
 			}
 
 			/* TODO: check return, locking me */
@@ -503,7 +517,6 @@ int mount_autofs_direct(struct autofs_point *ap)
 		map = map->next;
 	}
 	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
 
 	return 0;
 }
@@ -530,7 +543,9 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 			      me->key);
 			return 0;
 		}
-		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
+		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
+		if (ioctlfd == -1)
+			return 1;
 		opened = 1;
 	}
 
@@ -583,7 +598,7 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 
 	retries = UMOUNT_RETRIES;
 	while ((rv = umount(me->key)) == -1 && retries--) {
-		struct timespec tm = {0, 200000000};
+		struct timespec tm = {0, 50000000};
 		if (errno != EBUSY)
 			break;
 		nanosleep(&tm, NULL);
@@ -618,7 +633,7 @@ force_umount:
 	return rv;
 }
 
-int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *root, const char *offset)
+int mount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 {
 	const char *str_offset = mount_type_str(t_offset);
 	struct ioctl_ops *ops = get_ioctl_ops();
@@ -630,7 +645,7 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 	const char *hosts_map_name = "-hosts";
 	const char *map_name = hosts_map_name;
 	const char *type;
-	char mountpoint[PATH_MAX];
+	struct mnt_list *mnt;
 
 	if (ops->version && ap->flags & MOUNT_FLAG_REMOUNT) {
 		ret = try_remount(ap, me, t_offset);
@@ -644,6 +659,11 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			if (ap->state != ST_READMAP)
 				warn(ap->logopt,
 				     "trigger %s already mounted", me->key);
+			mnt = mnts_add_mount(ap, me->key, MNTS_OFFSET);
+			if (!mnt)
+				error(ap->logopt,
+				      "failed to add offset mount %s to mounted list",
+				      me->key);
 			return MOUNT_OFFSET_OK;
 		}
 
@@ -676,36 +696,14 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path, ap->kpipefd, str_offset);
+		mp->options = make_options_string(ap->path,
+				ap->kpipefd, str_offset, ap->flags);
 		if (!mp->options)
 			return MOUNT_OFFSET_OK;
-
-		if ((ap->flags & MOUNT_FLAG_STRICTEXPIRE) &&
-		    ((get_kver_major() == 5 && get_kver_minor() > 3) ||
-		     (get_kver_major() > 5))) {
-			char *tmp = realloc(mp->options, strlen(mp->options) + 12);
-			if (tmp) {
-				strcat(tmp, ",strictexpire");
-				mp->options = tmp;
-			}
-		}
-
-		if ((ap->flags & MOUNT_FLAG_IGNORE) &&
-		    ((get_kver_major() == 5 && get_kver_minor() > 4) ||
-		     (get_kver_major() > 5))) {
-			char *tmp = realloc(mp->options, strlen(mp->options) + 7);
-			if (tmp) {
-				strcat(tmp, ",ignore");
-				mp->options = tmp;
-			}
-		}
 	}
 
-	strcpy(mountpoint, root);
-	strcat(mountpoint, offset);
-
 	/* In case the directory doesn't exist, try to mkdir it */
-	if (mkdir_path(mountpoint, mp_mode) < 0) {
+	if (mkdir_path(me->key, mp_mode) < 0) {
 		if (errno == EEXIST) {
 			/*
 			 * If the mount point directory is a real mount
@@ -713,8 +711,8 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			 * a mount that has been automatically mounted by
 			 * the kernel NFS client.
 			 */
-			if (me->multi != me &&
-			    is_mounted(mountpoint, MNTS_REAL))
+			if (!IS_MM_ROOT(me) &&
+			    is_mounted(me->key, MNTS_REAL))
 				return MOUNT_OFFSET_IGNORE;
 
 			/* 
@@ -734,13 +732,13 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			debug(ap->logopt,
 			     "can't create mount directory: %s, %s",
-			     mountpoint, estr);
+			     me->key, estr);
 			return MOUNT_OFFSET_FAIL;
 		} else {
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			crit(ap->logopt,
 			     "failed to create mount directory: %s, %s",
-			     mountpoint, estr);
+			     me->key, estr);
 			return MOUNT_OFFSET_FAIL;
 		}
 	} else {
@@ -750,50 +748,51 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 
 	debug(ap->logopt,
 	      "calling mount -t autofs " SLOPPY " -o %s automount %s",
-	      mp->options, mountpoint);
+	      mp->options, me->key);
 
 	type = ap->entry->maps->type;
 	if (!type || strcmp(ap->entry->maps->type, "hosts"))
 		map_name = me->mc->map->argv[0];
 
-	ret = mount(map_name, mountpoint, "autofs", MS_MGC_VAL, mp->options);
+	ret = mount(map_name, me->key, "autofs", MS_MGC_VAL, mp->options);
 	if (ret) {
 		crit(ap->logopt,
 		     "failed to mount offset trigger %s at %s",
-		     me->key, mountpoint);
+		     me->key, me->key);
 		goto out_err;
 	}
 
-	ret = stat(mountpoint, &st);
+	ret = stat(me->key, &st);
 	if (ret == -1) {
-		error(ap->logopt,
-		     "failed to stat direct mount trigger %s", mountpoint);
-		goto out_umount;
-	}
+		int save_errno = errno;
 
-	ops->open(ap->logopt, &ioctlfd, st.st_dev, mountpoint);
-	if (ioctlfd < 0) {
-		crit(ap->logopt, "failed to create ioctl fd for %s", mountpoint);
-		goto out_umount;
+		error(ap->logopt,
+		     "failed to stat direct mount trigger %s", me->key);
+		if (save_errno != ENOENT)
+			goto out_umount;
+		goto out_err;
 	}
+	me->dev = st.st_dev;
+	me->ino = st.st_ino;
+
+	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
+	if (ioctlfd < 0)
+		goto out_umount;
 
 	ops->timeout(ap->logopt, ioctlfd, timeout);
-	cache_set_ino_index(me->mc, me->key, st.st_dev, st.st_ino);
-	if (ap->logopt & LOGOPT_DEBUG)
-		notify_mount_result(ap, mountpoint, timeout, str_offset);
-	else
-		notify_mount_result(ap, me->key, timeout, str_offset);
+	cache_set_ino_index(me->mc, me);
+	notify_mount_result(ap, me->key, timeout, str_offset);
 	ops->close(ap->logopt, ioctlfd);
 
-	debug(ap->logopt, "mounted trigger %s at %s", me->key, mountpoint);
+	debug(ap->logopt, "mounted trigger %s", me->key);
 
 	return MOUNT_OFFSET_OK;
 
 out_umount:
-	umount(mountpoint);
+	umount(me->key);
 out_err:
-	if (stat(mountpoint, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
-		 rmdir_path(ap, mountpoint, st.st_dev);
+	if (stat(me->key, &st) == 0 && me->flags & MOUNT_FLAG_DIR_CREATED)
+		 rmdir_path(ap, me->key, st.st_dev);
 
 	return MOUNT_OFFSET_FAIL;
 }
@@ -801,8 +800,8 @@ out_err:
 void *expire_proc_direct(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	struct mnt_list *mnts = NULL, *next;
-	struct list_head list, *p;
+	struct mnt_list *mnt;
+	LIST_HEAD(mnts);
 	struct expire_args *ea;
 	struct expire_args ec;
 	struct autofs_point *ap;
@@ -834,31 +833,24 @@ void *expire_proc_direct(void *arg)
 
 	left = 0;
 
-	mnts = tree_make_mnt_tree("/");
-	pthread_cleanup_push(mnts_cleanup, mnts);
-
-	/* Get a list of mounts select real ones and expire them if possible */
-	INIT_LIST_HEAD(&list);
-	if (!tree_get_mnt_list(mnts, &list, "/", 0)) {
-		ec.status = 0;
-		return NULL;
-	}
-
-	list_for_each(p, &list) {
-		next = list_entry(p, struct mnt_list, list);
-
+	/* Get the list of real mounts and expire them if possible */
+	mnts_get_expire_list(&mnts, ap);
+	if (list_empty(&mnts))
+		goto done;
+	pthread_cleanup_push(mnts_cleanup, &mnts);
+	list_for_each_entry(mnt, &mnts, expire) {
 		/*
 		 * All direct mounts must be present in the map
 		 * entry cache.
 		 */
 		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 		master_source_readlock(ap->entry);
-		me = lookup_source_mapent(ap, next->mp, LKP_DISTINCT);
+		me = lookup_source_mapent(ap, mnt->mp, LKP_DISTINCT);
 		pthread_cleanup_pop(1);
 		if (!me)
 			continue;
 
-		if (next->flags & MNTS_AUTOFS) {
+		if (mnt->flags & (MNTS_AUTOFS|MNTS_OFFSET)) {
 			struct stat st;
 			int ioctlfd;
 
@@ -869,8 +861,8 @@ void *expire_proc_direct(void *arg)
 			 * one of them and pass on state change.
 			 */
 			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-			if (next->flags & MNTS_INDIRECT) {
-				master_notify_submount(ap, next->mp, ap->state);
+			if (mnt->flags & MNTS_AUTOFS) {
+				master_notify_submount(ap, mnt->mp, ap->state);
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
@@ -880,11 +872,9 @@ void *expire_proc_direct(void *arg)
 				continue;
 			}
 
-			/* It's got a mount, deal with in the outer loop */
-			if (is_mounted(me->key, MNTS_REAL)) {
-				pthread_setcancelstate(cur_state, NULL);
-				continue;
-			}
+			/* It's got a mount, just send the expire. */
+			if (is_mounted(me->key, MNTS_REAL))
+				goto cont;
 
 			/*
 			 * Maybe a manual umount, repair.
@@ -901,19 +891,20 @@ void *expire_proc_direct(void *arg)
 			cache_writelock(me->mc);
 			if (me->ioctlfd != -1 && 
 			    fstat(me->ioctlfd, &st) != -1 &&
-			    !count_mounts(ap, next->mp, st.st_dev)) {
+			    !count_mounts(ap, mnt->mp, st.st_dev)) {
 				ops->close(ap->logopt, me->ioctlfd);
 				me->ioctlfd = -1;
 				cache_unlock(me->mc);
+				mnts_remove_mount(mnt->mp, MNTS_MOUNTED);
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
 			cache_unlock(me->mc);
-
+cont:
 			ioctlfd = me->ioctlfd;
 
-			ret = ops->expire(ap->logopt, ioctlfd, next->mp, how);
-			if (ret) {
+			ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
+			if (ret == 1) {
 				left++;
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
@@ -935,11 +926,11 @@ void *expire_proc_direct(void *arg)
 		if (ap->state == ST_EXPIRE || ap->state == ST_PRUNE)
 			pthread_testcancel();
 
-		debug(ap->logopt, "send expire to trigger %s", next->mp);
+		debug(ap->logopt, "send expire to trigger %s", mnt->mp);
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		ret = ops->expire(ap->logopt, ioctlfd, next->mp, how);
-		if (ret)
+		ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
+		if (ret == 1)
 			left++;
 		pthread_setcancelstate(cur_state, NULL);
 	}
@@ -947,7 +938,7 @@ void *expire_proc_direct(void *arg)
 
 	if (left)
 		debug(ap->logopt, "%d remaining in %s", left, ap->path);
-
+done:
 	ec.status = left;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -1056,13 +1047,18 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 		map = map->next;
 	}
 
-	if (!me) {
+	if (!me || me->len >= PATH_MAX) {
 		/*
 		 * Shouldn't happen as we have been sent this following
 		 * successful thread creation and lookup.
 		 */
-		crit(ap->logopt, "can't find map entry for (%lu,%lu)",
-		    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
+		if (!me)
+			crit(ap->logopt, "can't find map entry for (%lu,%lu)",
+			    (unsigned long) pkt->dev, (unsigned long) pkt->ino);
+		else {
+			cache_unlock(mc);
+			crit(ap->logopt, "lookup key is too long");
+		}
 		master_source_unlock(ap->entry);
 		pthread_setcancelstate(state, NULL);
 		return 1;
@@ -1071,9 +1067,9 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	/* Can't expire it if it isn't mounted */
 	if (me->ioctlfd == -1) {
 		int ioctlfd;
-		ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
+
+		ioctlfd = open_ioctlfd(ap, me->key, me->dev);
 		if (ioctlfd == -1) {
-			crit(ap->logopt, "can't open ioctlfd for %s", me->key);
 			cache_unlock(mc);
 			master_source_unlock(ap->entry);
 			pthread_setcancelstate(state, NULL);
@@ -1108,7 +1104,6 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	mt->ap = ap;
 	mt->ioctlfd = me->ioctlfd;
 	mt->mc = mc;
-	/* TODO: check length here */
 	strcpy(mt->name, me->key);
 	mt->dev = me->dev;
 	mt->type = NFY_EXPIRE;
@@ -1160,12 +1155,18 @@ int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_di
 	return 0;
 }
 
-static void mount_send_fail(void *arg)
+static void mount_send_status(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *mt = arg;
 	struct autofs_point *ap = mt->ap;
-	ops->send_fail(ap->logopt, mt->ioctlfd, mt->wait_queue_token, -ENOENT);
+
+	if (mt->status)
+		ops->send_fail(ap->logopt, mt->ioctlfd,
+			       mt->wait_queue_token, mt->status);
+	else
+		ops->send_ready(ap->logopt,
+				mt->ioctlfd, mt->wait_queue_token);
 	ops->close(ap->logopt, mt->ioctlfd);
 }
 
@@ -1174,6 +1175,7 @@ static void *do_mount_direct(void *arg)
 	struct ioctl_ops *ops = get_ioctl_ops();
 	struct pending_args *args, mt;
 	struct autofs_point *ap;
+	struct mapent *me;
 	struct stat st;
 	int status, state;
 
@@ -1194,7 +1196,8 @@ static void *do_mount_direct(void *arg)
 
 	pending_mutex_unlock(args);
 
-	pthread_cleanup_push(mount_send_fail, &mt);
+	mt.status = 0;
+	pthread_cleanup_push(mount_send_status, &mt);
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
@@ -1205,23 +1208,11 @@ static void *do_mount_direct(void *arg)
 	}
 
 	status = fstat(mt.ioctlfd, &st);
-	if (status == -1) {
-		error(ap->logopt,
-		      "can't stat direct mount trigger %s", mt.name);
-		ops->send_fail(ap->logopt,
-			       mt.ioctlfd, mt.wait_queue_token, -ENOENT);
-		ops->close(ap->logopt, mt.ioctlfd);
-		pthread_setcancelstate(state, NULL);
-		pthread_exit(NULL);
-	}
-
-	status = stat(mt.name, &st);
 	if (status != 0 || !S_ISDIR(st.st_mode) || st.st_dev != mt.dev) {
 		error(ap->logopt,
 		     "direct trigger not valid or already mounted %s",
 		     mt.name);
-		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
-		ops->close(ap->logopt, mt.ioctlfd);
+		mt.status = -EINVAL;
 		pthread_setcancelstate(state, NULL);
 		pthread_exit(NULL);
 	}
@@ -1239,14 +1230,17 @@ static void *do_mount_direct(void *arg)
 	 */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 	if (status) {
-		struct mapent *me;
+		struct mnt_list *sbmnt;
 		struct statfs fs;
 		unsigned int close_fd = 0;
+		unsigned int flags = MNTS_DIRECT|MNTS_MOUNTED;
 
+		sbmnt = mnts_find_submount(mt.name);
 		if (statfs(mt.name, &fs) == -1 ||
-		   (fs.f_type == AUTOFS_SUPER_MAGIC &&
-		    !master_find_submount(ap, mt.name)))
+		   (fs.f_type == AUTOFS_SUPER_MAGIC && !sbmnt))
 			close_fd = 1;
+		if (sbmnt)
+			mnts_put_mount(sbmnt);
 		cache_writelock(mt.mc);
 		if ((me = cache_lookup_distinct(mt.mc, mt.name))) {
 			/*
@@ -1254,22 +1248,38 @@ static void *do_mount_direct(void *arg)
 			 * for direct mount multi-mounts with no real mount at
 			 * their base so they will be expired.
 			 */
-			if (close_fd && me == me->multi)
+			if (close_fd && IS_MM_ROOT(me))
 				close_fd = 0;
 			if (!close_fd)
 				me->ioctlfd = mt.ioctlfd;
+			if (IS_MM(me) && !IS_MM_ROOT(me))
+				flags |= MNTS_OFFSET;
 		}
 		ops->send_ready(ap->logopt, mt.ioctlfd, mt.wait_queue_token);
 		cache_unlock(mt.mc);
 		if (close_fd)
 			ops->close(ap->logopt, mt.ioctlfd);
+
 		info(ap->logopt, "mounted %s", mt.name);
+
+		mnts_set_mounted_mount(ap, mt.name, flags);
+
+		conditional_alarm_add(ap, ap->exp_runfreq);
 	} else {
 		/* TODO: get mount return status from lookup_nss_mount */
 		ops->send_fail(ap->logopt,
 			       mt.ioctlfd, mt.wait_queue_token, -ENOENT);
 		ops->close(ap->logopt, mt.ioctlfd);
 		info(ap->logopt, "failed to mount %s", mt.name);
+
+		/* If this is a multi-mount subtree mount failure
+		 * ensure the tree continues to expire.
+		 */
+		cache_readlock(mt.mc);
+		me = cache_lookup_distinct(mt.mc, mt.name);
+		if (me && IS_MM(me) && !IS_MM_ROOT(me))
+			conditional_alarm_add(ap, ap->exp_runfreq);
+		cache_unlock(mt.mc);
 	}
 	pthread_setcancelstate(state, NULL);
 
@@ -1289,7 +1299,7 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 	char buf[MAX_ERR_BUF];
 	int status = 0;
 	struct timespec wait;
-	int ioctlfd, len, state;
+	int ioctlfd, state;
 	unsigned int kver_major = get_kver_major();
 	unsigned int kver_minor = get_kver_minor();
 
@@ -1352,8 +1362,8 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		close(me->ioctlfd);
 		me->ioctlfd = -1;
 	}
-	ops->open(ap->logopt, &ioctlfd, me->dev, me->key);
 
+	ioctlfd = open_ioctlfd(ap, me->key, me->dev);
 	if (ioctlfd == -1) {
 		cache_unlock(mc);
 		master_source_unlock(ap->entry);
@@ -1391,8 +1401,7 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 		return 0;
 	}
 
-	len = strlen(me->key);
-	if (len >= PATH_MAX) {
+	if (me->len >= PATH_MAX) {
 		error(ap->logopt, "direct mount path too long %s", me->key);
 		ops->send_fail(ap->logopt,
 			       ioctlfd, pkt->wait_queue_token, -ENAMETOOLONG);
@@ -1431,7 +1440,7 @@ int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_
 	mt->ioctlfd = ioctlfd;
 	mt->mc = mc;
 	strcpy(mt->name, me->key);
-	mt->len = len;
+	mt->len = me->len;
 	mt->dev = me->dev;
 	mt->type = NFY_MOUNT;
 	mt->uid = pkt->uid;
